html[data-theme="light"] {
    --pst-color-secondary: #6c757d;
}

html[data-theme="dark"] {
    --pst-color-secondary: #6c757d;
}

.version-switcher__container a[data-version-name*="stable"] {
    background-color: var(--pst-color-success);
}

.version-switcher__button[data-active-version-name*="dev"] {
    background-color: var(--pst-color-warning);
}
